//+------------------------------------------------------------------+
//|                                                     SignalCT.mqh |
//|                   Copyright 2009-2013, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#include <Expert\ExpertSignal.mqh>
#include                            <Math\Stat\stat.mqh>
#include                            <Math\Alglib\alglib.mqh>
#include                            <Math\Alglib\dataanalysis.mqh>
#define __PI 245850922/78256779
#define __PHI 0.61803398874989
#define __DECODERS 1
#define __IN_ROWS 1
#define __IN_COLS 4
#define __OUT_ROWS 1
#define __OUT_COLS 4
#define __HIDDEN __IN_COLS*__OUT_COLS
// wizard description start
//+------------------------------------------------------------------+
//| Description of the class                                         |
//| Title=Signals based on 'Category Theory' Transformer article 20r2|
//| Type=SignalAdvanced                                              |
//| Name=CategoryTheory                                              |
//| ShortName=CT                                                     |
//| Class=CSignalCT                                                  |
//| Page=signal_ct                                                   |
//| Parameter=Level,int,250,Level Gap                                |
//| Parameter=HiddenMultiple,int,1,Hidden Layer Multiple             |
//+------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| Class CSignalCT.                                                 |
//| Appointment: Class of generator of trade signals based on        |
//|               'Category Theory' Transformer article 20r2         |
//| Derives from class CExpertSignal.                                |
//+------------------------------------------------------------------+

struct Ssimilarity
  {
      matrix               queries,keys,values; 
   
                           Ssimilarity()
                           {
                              queries.Init(1,__IN_COLS);
                              keys.Init(1,__IN_COLS);
                              values.Init(1,__IN_COLS);
                           };
                           ~Ssimilarity(){};
  };

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CSignalCT : public CExpertSignal
  {
protected:
   //--- adjusted parameters
   int                     m_hidden;               // 
   int                     m_level;               // 
   double                  m_decoding;
   double                  m_training_benchmark;
   double                  m_benchmark;
   bool                    m_init;
   //

public:
   //--- methods of setting adjustable parameters
   void                    Level(int value)              { m_level=value;        }
   void                    HiddenMultiple(int value)     { m_hidden=value;       }
   void                    Decoding(double value)        { m_decoding=value;     }
   
   
   
   //--- method of verification of settings
   virtual bool      ValidationSettings(void);
   //--- method of creating the indicator and timeseries
   virtual bool      InitIndicators(CIndicators *indicators);
   //--- methods of checking if the market models are formed
   virtual int       LongCondition(void);
   virtual int       ShortCondition(void);
   
   virtual void      GetDirection(void);// override;
   //---
                     CSignalCT(void);
                    ~CSignalCT(void);
   //--- methods of initialization of protected data
   
   virtual vector          PositionalEncoding(int Positions);
   virtual matrix          DotProduct(matrix &Queries,matrix &Keys);
   virtual void            SoftMax(vector &Values,vector &Probabilities);
   virtual void            Decode(int &DecoderIndex,matrix &Input,matrix &Sum,matrix &Output);
   virtual void            FeedForward(int &DecoderIndex,matrix &Input,matrix &Output);
   virtual void            SelfAttention(int &DecoderIndex,matrix &Input,matrix &Queries,matrix &Keys,matrix &Values);
   
   CMLPTrain               m_train;
   CMLPBase                m_base_q[__DECODERS],m_base_k[__DECODERS],m_base_v[__DECODERS],m_base_f[__DECODERS];
   CMultilayerPerceptron   m_mlp_q[__DECODERS],m_mlp_k[__DECODERS],m_mlp_v[__DECODERS],m_mlp_f[__DECODERS];
   
   matrix                  m_sum_new,m_sum_old;
  };
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
CSignalCT::CSignalCT(void)
  {
//--- initialization of protected data
      m_used_series=USE_SERIES_TIME+USE_SERIES_SPREAD+USE_SERIES_OPEN+USE_SERIES_HIGH+USE_SERIES_LOW+USE_SERIES_CLOSE;
      
      m_training_benchmark=0.0;
      m_benchmark=m_account.Balance();
      
      for(int i=0;i<__DECODERS;i++)
      {
         m_base_q[i].MLPCreate1(__IN_COLS,__HIDDEN,__OUT_COLS,m_mlp_q[i]);
         m_base_k[i].MLPCreate1(__IN_COLS,__HIDDEN,__OUT_COLS,m_mlp_k[i]);
         m_base_v[i].MLPCreate1(__IN_COLS,__HIDDEN,__OUT_COLS,m_mlp_v[i]);
         m_base_f[i].MLPCreate1(__IN_COLS,__HIDDEN,__OUT_COLS,m_mlp_f[i]);
      }
         
      m_init=false;
  }
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
CSignalCT::~CSignalCT(void)
  {
      m_benchmark=fmax(m_benchmark,m_account.Balance());
      
  }
//+------------------------------------------------------------------+
//| Validation settings protected data.                              |
//+------------------------------------------------------------------+
bool CSignalCT::ValidationSettings(void)
  {
//--- validation settings of additional filters
   if(!CExpertSignal::ValidationSettings())
      return(false);
      
//--- validation settings of additional filters
   if(PeriodSeconds(PERIOD_W1)<=PeriodSeconds(m_period)){
      printf(__FUNCSIG__+" time frame should be less than weekly. ");
      return(false);}
      
//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| Create indicators.                                               |
//+------------------------------------------------------------------+
bool CSignalCT::InitIndicators(CIndicators *indicators)
  {
//--- check pointer
   if(indicators==NULL)
      return(false);
//--- initialization of indicators and timeseries of additional filters
   if(!CExpertSignal::InitIndicators(indicators))
      return(false);
      
      
      
//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| "Voting" that price will grow.                                   |
//+------------------------------------------------------------------+
int CSignalCT::LongCondition(void)
  {
      int result=0;
      
      GetDirection();
      //printf(__FUNCSIG__+" direction: "+DoubleToString(m_decoding,2));
      
      if(m_decoding<0)
      {
         result=0;
      }
      else if(m_decoding>0)
      {
         result=100;
      }
      
      return(result);
  }
//+------------------------------------------------------------------+
//| "Voting" that price will fall.                                   |
//+------------------------------------------------------------------+
int CSignalCT::ShortCondition(void)
  {
      int result=0;
      
      GetDirection();
      //printf(__FUNCSIG__+" direction: "+DoubleToString(m_decoding,2));
      
      if(m_decoding>0)
      {
         result=0;
      }
      else if(m_decoding<0)
      {
         result=100;
      }
      
      return(result);
  }
//+------------------------------------------------------------------+
//| Positional Encoding vector given length.                         |
//+------------------------------------------------------------------+
vector CSignalCT::PositionalEncoding(int Positions)
   {
      vector _positions;
      _positions.Init(Positions);_positions.Fill(0.0);
      
      for(int i=0;i<Positions;i++)
      {
         for(int ii=0;ii<Positions;ii++)
         {
            _positions[i]+=MathSin((((ii+1)/Positions)*(i+1))*__PI);
         }
      }
      
      return(_positions);
   }
//+------------------------------------------------------------------+
//| DotProduct for Query and Key Matrices Function.                  |
//+------------------------------------------------------------------+
matrix CSignalCT::DotProduct(matrix &Queries,matrix &Keys)
   {
      Queries.ReplaceNan(0.0);
      Keys.ReplaceNan(0.0);
      
      matrix _dot; _dot.Init(1,1); _dot.Fill(0.0);
      //
      if(int(Queries.Cols())==int(Keys.Rows()) && (int(Queries.Rows())==int(Keys.Cols())||int(Queries.Rows())==1||int(Keys.Cols())==1))
      {
         _dot.Init(int(Queries.Rows()),int(Keys.Cols())); _dot.Fill(0.0);
         //
         for(int i=0;i<int(Queries.Rows());i++)
         {
            for(int ii=0;ii<int(Keys.Cols());ii++)
            {
               for(int iii=0;iii<int(Keys.Rows());iii++)
               {
                  _dot[i][ii]+=(Queries[i][iii]*Keys[iii][ii]);
               }
            }
         }
      }
      else if(int(Queries.Rows())==int(Keys.Rows()) && int(Queries.Cols())==int(Keys.Cols()) && (int(Queries.Rows())==1||int(Keys.Cols())==1))
      {
         if(int(Queries.Rows())==1)
         {
            for(int i=0;i<int(Queries.Cols());i++)
            {
               _dot[0][0]+=(Queries[0][i]*Keys[0][i]);
            }
         }
         else if(int(Keys.Cols())==1)
         {
            for(int i=0;i<int(Keys.Rows());i++)
            {
               _dot[0][0]+=(Queries[i][0]*Keys[i][0]);
            }
         }
      }
      
      if(int(Keys.Cols())!=0 && MathIsValidNumber(Keys.Cols())){ _dot/=sqrt(Keys.Cols()); }
      
      _dot.ReplaceNan(0.0);
      
      return(_dot);
   }
//+------------------------------------------------------------------+
//| SoftMax Probability Distribution Function.                       |
//+------------------------------------------------------------------+
void CSignalCT::SoftMax(vector &Values,vector &Probabilities)
   {
      int _size=int(Values.Size());
      Probabilities.Init(_size);
      vector _e_values=MathExp(Values);
      double _sum=_e_values.Sum();
      
      if(_sum!=0.0 && MathIsValidNumber(_sum))
      {
         for(int s=0;s<_size;s++)
         {
            Probabilities[s]=_e_values[s]/_sum;
         }
      }
   }
//+------------------------------------------------------------------+
//| Feed Forward Function.                                           |
//+------------------------------------------------------------------+
void CSignalCT::FeedForward(int &DecoderIndex,matrix &Input,matrix &Output)
   {
      Input.ReplaceNan(0.0);
      //
      Output.Init(int(Input.Rows()),__OUT_COLS);Output.Fill(0.0);
      
      for(int i=0;i<int(Input.Rows());i++)
      {
         double _x_inputs[],_outputs[];
         vector _i=Input.Row(i);ArrayResize(_x_inputs,int(_i.Size()));
         for(int ii=0;ii<int(_i.Size());ii++){ _x_inputs[ii]=_i[ii]; }
         
         m_base_f[DecoderIndex].MLPProcess(m_mlp_f[DecoderIndex],_x_inputs,_outputs);
         
         for(int ii=0;ii<int(_outputs.Size());ii++){ if(!MathIsValidNumber(_outputs[ii])){ _outputs[ii]=0.0; }}
         
         for(int ii=0;ii<int(Output.Cols());ii++){ Output[i][ii]=_outputs[ii]; }
      }
   }
//+------------------------------------------------------------------+
//| Self Attention Function.                                         |
//+------------------------------------------------------------------+
void CSignalCT::SelfAttention(int &DecoderIndex,matrix &Input,matrix &Queries,matrix &Keys,matrix &Values)
   {
      Input.ReplaceNan(0.0);
      //
      Queries.Init(int(Input.Rows()),int(Input.Cols()));Queries.Fill(0.0);
      Keys.Init(int(Input.Rows()),int(Input.Cols()));Keys.Fill(0.0);
      Values.Init(int(Input.Rows()),int(Input.Cols()));Values.Fill(0.0);
      
      for(int i=0;i<int(Input.Rows());i++)
      {
         double _x_inputs[],_q_outputs[],_k_outputs[],_v_outputs[];
         vector _i=Input.Row(i);ArrayResize(_x_inputs,int(_i.Size()));
         for(int ii=0;ii<int(_i.Size());ii++){ _x_inputs[ii]=_i[ii]; }
         
         m_base_q[DecoderIndex].MLPProcess(m_mlp_q[DecoderIndex],_x_inputs,_q_outputs);
         m_base_k[DecoderIndex].MLPProcess(m_mlp_k[DecoderIndex],_x_inputs,_k_outputs);
         m_base_v[DecoderIndex].MLPProcess(m_mlp_v[DecoderIndex],_x_inputs,_v_outputs);
         
         for(int ii=0;ii<int(_q_outputs.Size());ii++){ if(!MathIsValidNumber(_q_outputs[ii])){ _q_outputs[ii]=0.0; }}
         for(int ii=0;ii<int(_k_outputs.Size());ii++){ if(!MathIsValidNumber(_k_outputs[ii])){ _k_outputs[ii]=0.0; }}
         for(int ii=0;ii<int(_v_outputs.Size());ii++){ if(!MathIsValidNumber(_v_outputs[ii])){ _v_outputs[ii]=0.0; }}
         
         for(int ii=0;ii<int(Queries.Cols());ii++){ Queries[i][ii]=_q_outputs[ii]; }   Queries.ReplaceNan(0.0);
         for(int ii=0;ii<int(Keys.Cols());ii++){ Keys[i][ii]=_k_outputs[ii]; }         Keys.ReplaceNan(0.0);
         for(int ii=0;ii<int(Values.Cols());ii++){ Values[i][ii]=_v_outputs[ii]; }     Values.ReplaceNan(0.0);
      }
   }
//+------------------------------------------------------------------+
//| Decode Function.                                                 |
//+------------------------------------------------------------------+
void CSignalCT::Decode(int &DecoderIndex,matrix &Input,matrix &Sum,matrix &Output)
   {
      Input.ReplaceNan(0.0);
      //
      //output matrices
      Sum.Init(1,int(Input.Cols()));Sum.Fill(0.0);
      
      Ssimilarity _s[];
      ArrayResize(_s,int(Input.Cols()));
      
      for(int i=int(Input.Rows())-1;i>=0;i--)
      {
         matrix _i;_i.Init(1,int(Input.Cols()));
         for(int ii=0;ii<int(Input.Cols());ii++)
         {
            _i[0][ii]=Input[i][ii];
         }
         //
         SelfAttention(DecoderIndex,_i,_s[i].queries,_s[i].keys,_s[i].values);
      }
      
      for(int i=int(Input.Cols())-1;i>=0;i--)
      {
         for(int ii=i;ii>=0;ii--)
         {
            matrix _similarity=DotProduct(_s[i].queries,_s[ii].keys);
            Sum+=DotProduct(_similarity,_s[i].values);
         }
      }
      //
      Sum.ReplaceNan(0.0);
      //
      FeedForward(DecoderIndex,Sum,Output);
   }
//+------------------------------------------------------------------+
//| Get Direction function.                                          |
//+------------------------------------------------------------------+
void CSignalCT::GetDirection()
   {
      matrix _i,_o[__DECODERS];
      
      m_open.Refresh(-1);
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      m_close.Refresh(-1);
      
      double _point=0.00001;
      if(StringFind(m_symbol.Name(),"JPY")>=0){ _point=0.001; }
      
      int _w=PeriodSeconds(PERIOD_W1)/PeriodSeconds(m_period);
      
      vector _position=PositionalEncoding(__IN_COLS);
   
      _i.Init(__IN_ROWS,__IN_COLS);
      
      for(int ii=0;ii<__IN_ROWS;ii++)
      {
         int _x=StartIndex()+ii;
         
         _i[ii][0]=(m_open.GetData(_x)-m_open.GetData(_x+_w))+_position[0];
         _i[ii][1]=(m_high.GetData(_x)-m_high.GetData(_x+_w))+_position[1];
         _i[ii][2]=(m_low.GetData(_x)-m_low.GetData(_x+_w))+_position[2];
         _i[ii][3]=(m_close.GetData(_x)-m_close.GetData(_x+_w))+_position[3];
      }
      
      for(int i=0;i<__DECODERS;i++)
      {
         _o[i].Init(__IN_ROWS,__IN_COLS);
         
         if(i==0)
         {
            Decode(i,_i,m_sum_new,_o[i]);
         }
         else
         {
            Decode(i,_o[i-1],m_sum_new,_o[i]);
         }
      }
      
      Decoding(_o[__DECODERS-1][0][3]);
      
      if(m_sum_old.Rows()==1 && m_sum_old.Cols()==_i.Cols())
      {
         int _info=0;
         CMLPReport _report;
         CMatrixDouble _xy;_xy.Resize(1,__IN_COLS+__OUT_COLS);
         
         _xy[0].Set(0,m_sum_old[0][0]); 
         _xy[0].Set(1,m_sum_old[0][1]); 
         _xy[0].Set(2,m_sum_old[0][2]); 
         _xy[0].Set(3,m_sum_old[0][3]);
         //
         _xy[0].Set(4,_i[0][0]);
         _xy[0].Set(5,_i[0][1]);
         _xy[0].Set(6,_i[0][2]);
         _xy[0].Set(7,_i[0][3]);
         //
         m_train.MLPTrainLM(m_mlp_f[__DECODERS-1],_xy,1,0.001,2,_info,_report);
      }
      
      m_sum_old=m_sum_new;
   }
//+------------------------------------------------------------------+
